/*

 Copyright 2013, 2014 SIANI - ULPGC
 Jose Juan Hernandez Cabrera
 Jose Evora Gomez
 Johan Sebastian Cortes Montenegro

 This File is Part of JavaFMI Project

 JavaFMI Project is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 JavaFMI Project is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with JavaFMI Library.  If not, see <http://www.gnu.org/licenses/>.

 */
package org.javafmi;

import org.javafmi.kernel.FmuProxy;
import org.javafmi.kernel.ProxyFactory;
import org.javafmi.kernel.Status;
import org.javafmi.kernel.operatingsystem.FmuFile;

import java.io.File;

public class Simulation {

    private static final int DEFAULT_TIMEOUT = 1000;
    private VariableWriter variableWriter;
    private VariableReader variableReader;
    private FmuProxy fmuProxy;
    private String proxyName;
    private String[] variables;
    private double simulationTime;

    protected Simulation() {
    }

    public Simulation(String fmuPath) {
        this(fmuPath, DEFAULT_TIMEOUT);
    }

    public Simulation(String fmuPath, double timeout) {
        this.fmuProxy = ProxyFactory.createProxyFor(new FmuFile(new File(fmuPath)));
        this.proxyName = fmuProxy.getModelName();
        this.variables = fmuProxy.getModelVariablesNames();
        this.fmuProxy.instantiate(timeout);
        this.variableWriter = new VariableWriter(fmuProxy);
        this.variableReader = new VariableReader(fmuProxy);
    }

    public Status init(double startTime) {
        return init(startTime, Double.MAX_VALUE);
    }

    public Status init(double startTime, double stopTime) {
        simulationTime = startTime;
        return fmuProxy.initialize(startTime, stopTime);
    }

    public Status doStep(double stepSize) {
        Status doStepStatus = fmuProxy.doStep(simulationTime, stepSize);
        if(doStepStatus == Status.OK) simulationTime += stepSize;
        return doStepStatus;
    }

    public Status cancelStep() {
        return fmuProxy.cancelStep();
    }

    public Status terminate() {
        Status terminateStatus = fmuProxy.terminate();
        fmuProxy.freeInstance();
        return terminateStatus;
    }

    public Status reset() {
        return fmuProxy.reset();
    }

    public Variable readVariable(String variableName) {
        return variableReader.read(variableName);
    }

    public Status writeVariable(String name, Object value) {
        return variableWriter.write(name, value);
    }

    public String[] listVariables() {
        return variables;
    }

    public String getInternalName() {
        return proxyName;
    }

    public boolean isTerminated() {
        return fmuProxy.isTerminated();
    }

    public double getSimulationTime() {
        return simulationTime;
    }

    protected static FmuProxy getProxy(Simulation simulation) {
        return simulation.fmuProxy;
    }
}
