/*

 Copyright 2013, 2014 SIANI - ULPGC
 Jose Juan Hernandez Cabrera
 Jose Evora Gomez
 Johan Sebastian Cortes Montenegro

 This File is Part of JavaFMI Project

 JavaFMI Project is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 JavaFMI Project is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with JavaFMI Library.  If not, see <http://www.gnu.org/licenses/>.

 */
package org.javafmi.kernel.v2rc2;

import org.javafmi.exceptions.MethodExecutionException;
import org.javafmi.exceptions.UnsupportedOperationException;
import org.javafmi.kernel.FmuProxyState;
import org.javafmi.kernel.State;
import org.javafmi.kernel.Status;
import org.javafmi.kernel.modeldescription.FmiVersion;
import org.javafmi.kernel.modeldescription.ModelDescriptionDeserializer;
import org.javafmi.kernel.modeldescription.ScalarVariable;
import org.javafmi.kernel.operatingsystem.FmuFile;
import org.javafmi.kernel.v2rc2.modeldescription.EnumerationType;
import org.javafmi.kernel.v2rc2.modeldescription.ModelDescription;
import org.javafmi.kernel.v2rc2.nativeaccess.FmiV2RC2;
import org.javafmi.kernel.v2rc2.nativeaccess.FmiV2RC2Interface;

import java.util.ArrayList;
import java.util.List;

import static org.javafmi.kernel.FmuProxyState.*;
import static org.javafmi.kernel.Status.*;

public class FmuProxy implements org.javafmi.kernel.FmuProxy {

    private FmuProxyState fmuProxyState;
    private ModelDescription modelDescription;
    private FmiV2RC2 library;
    private FmuFile fmuFile;
    private PendingVariablesList pendingVariables;

    public FmuProxy(FmuFile fmuFile) {
        this.modelDescription = (ModelDescription) ModelDescriptionDeserializer.deserialize(ModelDescription.class, fmuFile.getModelDescriptionFile());
        this.modelDescription.buildAttributes();
        this.fmuFile = fmuFile;
        this.pendingVariables = new PendingVariablesList();
    }

    @Override
    public String getModelName() {
        return modelDescription.getModelName();
    }

    @Override
    public FmiVersion getVersion() {
        return FmiVersion.TWO;
    }

    @Override
    public void instantiate(double timeout) {
        this.library = new FmiV2RC2(fmuFile.getLibraryPath(modelDescription.getModelIdentifier()));
        this.library.instantiate(
                modelDescription.getModelName(),
                FmiV2RC2Interface.FMI_CO_SIMULATION,
                modelDescription.getGuid(),
                "file:///" + fmuFile.getWorkingDirectory());
        this.fmuProxyState = INSTANTIATED;
    }

    @Override
    public void freeInstance() {
        library.freeInstance();
        fmuFile.deleteTemporal();
    }

    public Status setDebugLogging(boolean loggingOn, String[] categories) {
        return Status.translateStatus(library.setDebugLogging(loggingOn, categories));
    }

    @Override
    public Status initialize(double startTime, double stopTime) {
        if(fmuProxyState.equals(INITED)) return WARNING;
        if(ERROR == setUpExperiment(false, 0.01, startTime, false, stopTime))
            throw new MethodExecutionException("fmi2SetupExperiment: " + modelDescription.getModelIdentifier());
        if(ERROR == enterInitializationMode())
            throw new MethodExecutionException("fmi2SetupExperiment: " + modelDescription.getModelIdentifier());
        if(ERROR == setPendingVariables())
            throw new MethodExecutionException("fmi2SetupExperiment: " + modelDescription.getModelIdentifier());
        if(ERROR == exitInitializationMode())
            throw new MethodExecutionException("fmi2SetupExperiment: " + modelDescription.getModelIdentifier());
        return OK;
    }

    private Status setPendingVariables() {
        return pendingVariables.setAll(this);
    }

    public Status setUpExperiment(boolean toleranceDefined, double tolerance, double startTime, boolean stopTimeDefined, double stopTime) {
        return Status.translateStatus(library.setupExperiment(toleranceDefined, tolerance, startTime, stopTimeDefined, stopTime));
    }

    public Status enterInitializationMode() {
        return Status.translateStatus(library.enterInitializationMode());
    }

    public Status exitInitializationMode() {
        Status status = Status.translateStatus(library.exitInitializationMode());
        if(status == OK) fmuProxyState = INITED;
        return status;
    }

    @Override
    public Status terminate() {
        Status terminateStatus = Status.translateStatus(library.terminate());
        if(terminateStatus.equals(OK)) fmuProxyState = TERMINATED;
        else fmuProxyState = INITED;
        return terminateStatus;
    }

    @Override
    public Status reset() {
        Status resetStatus = Status.translateStatus(library.reset());
        if(resetStatus.equals(Status.OK)) fmuProxyState = INSTANTIATED;
        return resetStatus;
    }

    @Override
    public double getReal(Integer valueReference) {
        return library.getReal(new int[]{valueReference});
    }

    @Override
    public int getInteger(Integer valueReference) {
        return library.getInteger(new int[]{valueReference});
    }

    @Override
    public boolean getBoolean(Integer valueReference) {
        return library.getBoolean(new int[]{valueReference});
    }

    @Override
    public String getString(Integer valueReference) {
        return library.getString(new int[]{valueReference});
    }

    @Override
    public Object getEnumeration(ScalarVariable modelVariable) {
        EnumerationType enumeration = (EnumerationType) modelVariable.getType();
        Integer enumerationIndex;
        EnumerationType enumerationType = (EnumerationType) modelDescription.getTypeFromTypeDefinition(enumeration.getDeclaredType());
        if(modelVariable.getVariability().equalsIgnoreCase("constant")) {
            enumerationIndex = enumeration.getStart();
            return enumerationType.getItems().get(enumerationIndex - 1).getName();
        }
        enumerationIndex = getInteger(modelVariable.getValueReference());
        return enumerationType.getItems().get(enumerationIndex - 1).getName();
    }

    @Override
    public Status setReal(Integer valueReference, Double doubleValue) {
        if((fmuProxyState.equals(FmuProxyState.INSTANTIATED))) {
            pendingVariables.addReal(valueReference, doubleValue);
            return OK;
        }
        return Status.translateStatus(library.setReal(new int[]{valueReference}, new double[]{doubleValue}));
    }

    @Override
    public Status setInteger(Integer valueReference, Integer intValue) {
        if((fmuProxyState.equals(FmuProxyState.INSTANTIATED))) {
            pendingVariables.addInteger(valueReference, intValue);
            return OK;
        }
        return Status.translateStatus(library.setInteger(new int[]{valueReference}, new int[]{intValue}));
    }

    @Override
    public Status setBoolean(Integer valueReference, Boolean booleanValue) {
        if((fmuProxyState.equals(FmuProxyState.INSTANTIATED))) {
            pendingVariables.addBoolean(valueReference, booleanValue);
            return OK;
        }
        return Status.translateStatus(library.setBoolean(new int[]{valueReference}, new boolean[]{booleanValue}));
    }

    @Override
    public Status setString(Integer valueReference, String stringValue) {
        if((fmuProxyState.equals(FmuProxyState.INSTANTIATED))) {
            pendingVariables.addString(valueReference, stringValue);
            return OK;
        }
        return Status.translateStatus(library.setString(new int[]{valueReference}, new String[]{stringValue}));
    }

    public boolean canGetAndSetFmuState() {
        return modelDescription.getCoSimulation().canGetAndSetFmuState();
    }

    public boolean canSerializeState() {
        return modelDescription.getCoSimulation().canSerializeFMUState();
    }

    public State getState(State state) {
        if(!canGetAndSetFmuState())
            throw new UnsupportedOperationException("getState: " + modelDescription.getModelIdentifier());
        return state.setState(library.getFMUState(state.getState()));
    }

    public Status setState(State state) {
        if(!canGetAndSetFmuState())
            throw new UnsupportedOperationException("setState: " + modelDescription.getModelIdentifier());
        return Status.translateStatus(library.setFMUState(state.getState()));
    }

    public Status freeState(State state) {
        if(!canGetAndSetFmuState())
            throw new UnsupportedOperationException("freeState: " + modelDescription.getModelIdentifier());
        return Status.translateStatus(library.freeFMUState(state.getState()));
    }

    public int getStateSize(State state) {
        if(!canGetAndSetFmuState() || !canSerializeState())
            throw new UnsupportedOperationException("getStateSize: " + modelDescription.getModelIdentifier());
        return library.serializedFMUStateSize(state.getState());
    }

    public byte[] serializeState(State state) {
        return library.serializeFMUState(state.getState());
    }

    public State deserializeState(byte[] serializedState) {
        return State.newStateFrom(library.deSerializeFMUState(serializedState));
    }

    public double[] getDirectionalDerivative(int[] unknownVariablesValueReferences, int[] knownVariablesValueReferences, double[] knownDifferential) {
        if(!modelDescription.getCoSimulation().providesDirectionalDerivative())
            throw new UnsupportedOperationException("getDirectionalDerivative: " + modelDescription.getModelIdentifier());
        return library.getDirectionalDerivative(unknownVariablesValueReferences, knownVariablesValueReferences, knownDifferential);
    }

    @Override
    public Status doStep(double communicationPoint, double stepSize) {
        return Status.translateStatus(library.doStep(communicationPoint, stepSize, false));
    }

    @Override
    public Status cancelStep() {
        return Status.translateStatus(library.cancelStep());
    }

    @Override
    public ScalarVariable[] getModelVariables() {
        return modelDescription.getModelVariables();
    }

    @Override
    public ScalarVariable getVariable(String name) {
        return modelDescription.getScalarVariable(name);
    }

    @Override
    public boolean isTerminated() {
        return fmuProxyState.equals(TERMINATED);
    }

    @Override
    public String[] getModelVariablesNames() {
        List<String> variables = new ArrayList<>();
        for(ScalarVariable scalarVariable : modelDescription.getModelVariables()) {
            variables.add(scalarVariable.getName());
        }
        return variables.toArray(new String[variables.size()]);
    }
}
